
from asyncio import Event
from distutils import filelist
from typing import Callable, List, Tuple, Optional, AsyncGenerator
import os
import importlib.util
import uuid

from reactivestreams.publisher import Publisher
from reactivestreams.subscriber import Subscriber
from reactivestreams.subscription import Subscription
from rsocket.streams.stream_from_async_generator import StreamFromAsyncGenerator
from rsocket.fragment import Fragment
from rsocket.payload import Payload

from appbridge.progress_bar import TranferBar
from appbridge.prototol import Msg, Action, RequestFilesCallable
from appbridge import common
from appbridge.config import Config


#
# client request get files from me
#
# get <file_path_on_service>...
#
async def handle_get_action(action: str, params: List[str], 
                            session_dir: str, request_files: RequestFilesCallable) -> AsyncGenerator[bytes, None]:
    for p in params:
        full_path = os.path.join(Config.files_dir, p)
        norm_path = os.path.normpath(full_path)
        basename = os.path.basename(norm_path)
        dir = os.path.dirname(norm_path)

        # send file to client
        async for e in common.file_generator(dir, basename):
            yield e


#
# client request send files to me
#
# send <file_path_on_client>...
#
async def handle_send_action(action: str, params: List[str], 
                            session_dir: str, request_files: RequestFilesCallable) -> AsyncGenerator[bytes, None]:
    # request files from client
    async for e in request_files(params):
        print('request file msg', e)
        yield e


class ChannelSubscriber(Subscriber):

    file_received_size: int = 0
    file_total_size: int = 0
    file = None

    def __init__(self, action, params, session_dir: str, got_file_event: Event ) -> None:
        super().__init__()
        self.action = action
        self.params = params
        self.session_dir = session_dir
        self.got_file_event = got_file_event

        if not os.path.isdir(session_dir):
            os.makedirs(session_dir, 0o755)
            print('create session_dir: %s' % session_dir)
        
        print('[+] Channel subscriber init')


    def __del__(self):
        if self.file:
            self.file.close()
        if os.path.isdir(self.session_dir):
            os.rmdir(self.session_dir)
            print('delete session_dir: %s' % self.session_dir)

        print('[-] Channel subscriber deleted')

    def finish(self):
        self.__del__()

    def on_subscribe(self, subscription: Subscription):
        self.subscription = subscription
        subscription.request(1)

    def on_error(self, exception: Exception):
        print('Error from client on channel' + str(exception))
        self.finish()

    def on_complete(self):
        print('Completed from client on channel')
        self.finish()

    def on_next(self, value: Payload, is_complete=False):

        # print("From client: len=%d, complete=%s" % (len(value.data), "Yes" if is_complete else "No"))

        # complete signal
        datalen = len(value.data)
        if datalen == 0 and is_complete:
            print("From clinet: complete empty msg")
            self.finish()
            return

        # handle file receiving
        if self.file_received_size < self.file_total_size:
            self.file.write(value.data)
            self.file_received_size += datalen
            self.subscription.request(1)
            self.bar.next(datalen)
            return

        # handle msg
        msg = Msg.fromBytes(value.data)
        print('\nMsg: %s' % msg.toString())

        if msg.action == Action.ResponseOneFileInfo:
            (self.file, self.file_total_size) = common.prepare_receive_file(msg, Config.files_dir)
            self.file_received_size = 0
            self.bar = TranferBar('Receiving', max=self.file_total_size)
            print("Prepared to receive a file!")

        elif msg.action == Action.ResponseOneFileComplete:
            self.file.close()
            self.file = None
            self.file_received_size = 0
            self.file_total_size = 0
            self.bar.finish()
            print("Received a file!")

        if msg.action == Action.ResponseAllFilesComplete:
            self.got_file_event.set()
        
        if is_complete:
            self.finish()
        else:
            self.subscription.request(1)


    def publisher(self) -> Publisher:

        async def request_files(file_list) -> AsyncGenerator[bytes, None]:

            self.got_file_event.clear()

            # do request files
            req = Msg(Action.RequestFile, file_list)
            yield req.toBytes()

            # wait for file received
            await self.got_file_event.wait()


        async def generator() -> AsyncGenerator[Tuple[Fragment, bool], None]:

            # find an action handler to handle the action
            action = self.action
            module_file_path = Config.service_dir + '/' + action + '.py'
            module_name = action
            handle_action = None
            if Config.service_dir == '' or not os.path.isfile(module_file_path):
                if Config.files_dir != '':
                    if action == Action.GetFile:
                        handle_action = handle_get_action
                    elif action == Action.SendFile:
                        handle_action = handle_send_action      
            else:
                print("load %s from %s" % (module_name, module_file_path))
                module_spec = importlib.util.spec_from_file_location(module_name , module_file_path)
                module = importlib.util.module_from_spec(module_spec)
                module_spec.loader.exec_module(module)
                handle_action = module.handle_action

            if handle_action is None:
                raise Exception("Server don't support action \"%s\"" % action)
 
            async for data in handle_action(action, self.params, self.session_dir, request_files):
                yield Fragment(data), False
    
            # complete
            msg = Msg(Action.Complete, [])
            yield Fragment(msg.toBytes()), True

        return StreamFromAsyncGenerator(generator)


def entry(action, params: List[str]) -> Tuple[Optional[Publisher], Optional[Subscriber]]:

    id = uuid.uuid1()
    session_dir = os.path.join(Config.service_dir, 'sessions', str(id)) 

    request_file_event = Event()
    subscriber = ChannelSubscriber(action, params, session_dir, request_file_event)

    return subscriber.publisher(), subscriber
