
import asyncio
from asyncio import Event
from typing import AsyncGenerator, Tuple, Optional

from reactivestreams.publisher import Publisher
from reactivestreams.subscriber import Subscriber
from reactivestreams.subscription import Subscription
from rsocket.extensions.helpers import route, composite, authenticate_simple
from rsocket.extensions.mimetypes import WellKnownMimeTypes
from rsocket.fragment import Fragment
from rsocket.helpers import single_transport_provider
from rsocket.payload import Payload
from rsocket.rsocket_client import RSocketClient
from rsocket.streams.stream_from_async_generator import StreamFromAsyncGenerator
from rsocket.transports.tcp import TransportTCP

from rsocket.transports.aiohttp_websocket import websocket_client
from rsocket.rsocket_client import RSocketClient

from appbridge import bonjour, common
from appbridge.prototol import Msg, Action
from appbridge.progress_bar import TranferBar

def get_service_ip_port(service_name: str) -> Optional[Tuple[str, str]]: 
    info = bonjour.get_service_info(service_name)
    if info is None:
        print("can't find service %s" % service_name)
        return None
    (ip, port) = bonjour.get_ip_port(info)
    print("service address: %s%d" % ('ws://' + ip + ':', port))

    return (ip, port)



class ChannelSubscriber(Subscriber):

    file_received_size: int = 0
    file_total_size: int = 0
    file = None

    def __init__(self, completion_event: Event, notify_publisher_event: Event) -> None:
        super().__init__()
        self.completion_event = completion_event
        self.send_file_event = notify_publisher_event
        self.finished = False

    def finish(self):
        self.finished = True
        self.send_file_event.set()
        self.completion_event.set()
        if self.file is not None:
            self.file.close()
        print("subscriber finished!")

    def on_subscribe(self, subscription: Subscription):
        print('on_subscribe')
        self.subscription = subscription

    def on_error(self, exception: Exception):
        if self.finished:
            print('Skip error from server: ' + str(exception))
            return
        print('!!! Got Error from server: ' + str(exception))
        self.finish()

    def on_complete(self):
        if self.finished:
            print('Skip completed from server')
            return
        print('!!! Got Completed from server')
        self.finish()

    def on_next(self, value: Payload, is_complete=False):
        # print("Payload from server: len=%d, complete=%s" % (len(value.data), "Yes" if is_complete else "No"))

        datalen = len(value.data)

        # complete signal
        if datalen == 0 and is_complete:
            print("Skip complete empty msg")
            return

        # handle file receiving
        if self.file_received_size < self.file_total_size:
            self.file.write(value.data)
            self.file_received_size += datalen
            self.subscription.request(1)
            self.bar.next(datalen)
            return

        # handle msg   
        msg = Msg.fromBytes(value.data)
        print('\nMsg: %s' % msg.toString())

        if msg.action == Action.Complete:
            self.finish()

        elif msg.action == Action.RequestFile:
            self.filepath = msg.params[0]
            self.send_file_event.set()
        
        elif msg.action == Action.ResponseOneFileInfo:
            (self.file, self.file_total_size) = common.prepare_receive_file(msg, '.')
            self.file_received_size = 0
            self.bar = TranferBar('Receiving', max=self.file_total_size)
            print("Prepared to receive a file!")

        elif msg.action == Action.ResponseOneFileComplete:
            self.file.close()
            self.file = None
            self.file_received_size = 0
            self.file_total_size = 0
            self.bar.finish()
            print("Received a file!")

        self.subscription.request(1)

 
    def publisher(self) -> Publisher:
        async def generator() -> AsyncGenerator[Tuple[Fragment, bool], None]:
            while True:
                await self.send_file_event.wait()
                self.send_file_event.clear()
                if self.finished:
                    break
                async for e in  common.file_generator('.', self.filepath):
                    yield Fragment(e), False

            print("publisher finished!")

        return StreamFromAsyncGenerator(generator)


async def do_request(service_name, msg: Msg):
    (ip, port) = get_service_ip_port(service_name)
    print("reqest data: %s" % msg.toString())
    async with websocket_client('http://%s:%d' % (ip, port)) as client:
        channel_completion_event = Event()
        request_file_event = Event()

        subscriber = ChannelSubscriber(channel_completion_event, request_file_event)
        publisher = subscriber.publisher()

        requested = client.request_channel(Payload(msg.toBytes()), publisher)
        requested.initial_request_n(1).subscribe(subscriber)

        await channel_completion_event.wait()
        print('channel done!')

    print('do_request done!')



async def do_request_info(service_name):
    (ip, port) = get_service_ip_port(service_name)
    msg = Msg(Action.GetInfo, [])
    print("reqest data: %s" % msg.toString())
    async with websocket_client('http://%s:%d' % (ip, port)) as client:
        result = await client.request_response(Payload(msg.toBytes()))
        ret = Msg.fromBytes(result.data)
        print("Service Supported Actions:")
        for item in ret.params:
            print("\t%s" % item)

def request(service_name, action, params):
    msg = Msg(action, params)
    asyncio.run(do_request(service_name, msg))
    print("request done")

def request_info(service_name):
    asyncio.run(do_request_info(service_name))
    print("request done")